@@ -62,6 +62,11 @@ struct App
     GtkWidget      *area;
     gboolean	    ignore_gui_changes;
     GConfClient	   *client;
+
+    /* These are used while we are waiting for the ApplyConfiguration method to be executed over D-bus */
+    DBusGConnection *connection;
+    DBusGProxy *proxy;
+    DBusGProxyCall *proxy_call;
 };
 
 static void rebuild_gui (App *app);
@@ -69,6 +74,7 @@ static void on_rate_changed (GtkComboBox *box, gpointer data);
 static gboolean output_overlaps (GnomeOutputInfo *output, GnomeRRConfig *config);
 static void select_current_output_from_dialog_position (App *app);
 static void monitor_on_off_toggled_cb (GtkToggleButton *toggle, gpointer data);
+static void get_geometry (GnomeOutputInfo *output, int *w, int *h);
 
 static void
 error_message (App *app, const char *primary_text, const char *secondary_text)
@@ -76,7 +82,7 @@ error_message (App *app, const char *primary_text, const char *secondary_text)
     GtkWidget *dialog;
 
     dialog = gtk_message_dialog_new ((app && app->dialog) ? GTK_WINDOW (app->dialog) : NULL,
-				     GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_NO_SEPARATOR,
+				     GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
 				     GTK_MESSAGE_ERROR,
 				     GTK_BUTTONS_CLOSE,
 				     "%s", primary_text);
@@ -688,16 +694,64 @@ monitor_on_off_toggled_cb (GtkToggleButton *toggle, gpointer data)
     foo_scroll_area_invalidate (FOO_SCROLL_AREA (app->area));
 }
 
+static void
+realign_outputs_after_resolution_change (App *app, GnomeOutputInfo *output_that_changed, int old_width, int old_height)
+{
+    /* We find the outputs that were below or to the right of the output that
+     * changed, and realign them; we also do that for outputs that shared the
+     * right/bottom edges with the output that changed.  The outputs that are
+     * above or to the left of that output don't need to change.
+     */
+
+    int i;
+    int old_right_edge, old_bottom_edge;
+    int dx, dy;
+
+    g_assert (app->current_configuration != NULL);
+
+    old_right_edge = output_that_changed->x + old_width;
+    old_bottom_edge = output_that_changed->y + old_height;
+
+    dx = output_that_changed->width - old_width;
+    dy = output_that_changed->height - old_height;
+
+    for (i = 0; app->current_configuration->outputs[i] != NULL; i++) {
+	GnomeOutputInfo *output;
+	int output_width, output_height;
+
+	output = app->current_configuration->outputs[i];
+
+	if (output == output_that_changed || !output->connected)
+	    continue;
+
+	get_geometry (output, &output_width, &output_height);
+
+	if (output->x >= old_right_edge)
+	    output->x += dx;
+	else if (output->x + output_width == old_right_edge)
+	    output->x = output_that_changed->x + output_that_changed->width - output_width;
+
+	if (output->y >= old_bottom_edge)
+	    output->y += dy;
+	else if (output->y + output_height == old_bottom_edge)
+	    output->y = output_that_changed->y + output_that_changed->height - output_height;
+    }
+}
+
 static void
 on_resolution_changed (GtkComboBox *box, gpointer data)
 {
     App *app = data;
+    int old_width, old_height;
     int width;
     int height;
 
     if (!app->current_output)
 	return;
 
+    old_width = app->current_output->width;
+    old_height = app->current_output->height;
+
     if (get_mode (app->resolution_combo, &width, &height, NULL, NULL))
     {
 	app->current_output->width = width;
@@ -709,23 +763,7 @@ on_resolution_changed (GtkComboBox *box, gpointer data)
 	    app->current_output->on = TRUE;
     }
 
-#if 0
-    if (app->current_configuration)
-    {
-	x = 0;
-	for (i = 0; app->current_configuration->outputs[i] != NULL; ++i)
-	{
-	    GnomeOutputInfo *output = app->current_configuration->outputs[i];
-
-	    if (output->connected)
-	    {
-		output->x = x;
-
-		x += output->width;
-	    }
-	}
-    }
-#endif
+    realign_outputs_after_resolution_change (app, app->current_output, old_width, old_height);
 
     rebuild_rate_combo (app);
     rebuild_rotation_combo (app);
@@ -1541,7 +1579,7 @@ paint_output (App *app, cairo_t *cr, int i)
     cairo_move_to (cr, extent.x, extent.y);
 
     if (output->on)
-	cairo_set_source_rgb (cr, 0.2, 0.2, 0.8);
+	cairo_set_source_rgb (cr, 0.0, 0.0, 0.0);
     else
 	cairo_set_source_rgb (cr, 1.0, 1.0, 1.0);
 
@@ -1672,12 +1710,40 @@ check_required_virtual_size (App *app)
     }
 }
 
+/* Callback for dbus_g_proxy_begin_call() */
+static void
+apply_configuration_returned_cb (DBusGProxy       *proxy,
+				 DBusGProxyCall   *call_id,
+				 void             *data)
+{
+    App *app = data;
+    gboolean success;
+    GError *error;
+
+    g_assert (call_id == app->proxy_call);
+
+    error = NULL;
+    success = dbus_g_proxy_end_call (proxy, call_id, &error, G_TYPE_INVALID);
+
+    if (!success) {
+	error_message (app, _("Could not apply the selected configuration"), error->message);
+	g_error_free (error);
+    }
+
+    g_object_unref (app->proxy);
+    app->proxy = NULL;
+
+    dbus_g_connection_unref (app->connection);
+    app->connection = NULL;
+    app->proxy_call = NULL;
+
+    gtk_widget_set_sensitive (app->dialog, TRUE);
+}
+
 static void
 apply (App *app)
 {
     GError *error = NULL;
-    DBusGConnection *connection;
-    DBusGProxy *proxy;
 
     gnome_rr_config_sanitize (app->current_configuration);
 
@@ -1692,30 +1758,34 @@ apply (App *app)
 	return;
     }
 
-    connection = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
-    if (connection == NULL) {
+    g_assert (app->connection == NULL);
+    g_assert (app->proxy == NULL);
+    g_assert (app->proxy_call == NULL);
+
+    app->connection = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
+    if (app->connection == NULL) {
 	error_message (app, _("Could not get session bus while applying display configuration"), error->message);
 	g_error_free (error);
 	return;
     }
 
-    proxy = dbus_g_proxy_new_for_name (connection,
-				       "org.gnome.SettingsDaemon",
-				       "/org/gnome/SettingsDaemon/XRANDR",
-				       "org.gnome.SettingsDaemon.XRANDR");
-    if (!proxy) {
+    app->proxy = dbus_g_proxy_new_for_name (app->connection,
+					    "org.gnome.SettingsDaemon",
+					    "/org/gnome/SettingsDaemon/XRANDR",
+					    "org.gnome.SettingsDaemon.XRANDR");
+    if (!app->proxy) {
 	error_message (app, _("Could not get org.gnome.SettingsDaemon.XRANDR"), NULL);
+	dbus_g_connection_unref (app->connection);
+	app->connection = NULL;
 	return;
-
     }
 
-    if (!dbus_g_proxy_call (proxy, "ApplyConfiguration", &error, G_TYPE_INVALID, G_TYPE_INVALID)) {
-	error_message (app, _("Could not apply the selected configuration"), error->message);
-	g_error_free (error);
-    }
-
-    g_object_unref (proxy);
-    dbus_g_connection_unref (connection);
+    gtk_widget_set_sensitive (app->dialog, FALSE);
+    app->proxy_call = dbus_g_proxy_begin_call (app->proxy, "ApplyConfiguration",
+					       apply_configuration_returned_cb, app,
+					       NULL,
+					       G_TYPE_INVALID,
+					       G_TYPE_INVALID);
 }
 
 #if 0
